@@ -9,20 +9,12 @@
 #![doc(html_root_url = "https://docs.rs/wasm-bindgen/0.2")]
 #![cfg_attr(feature = "nightly", feature(unsize))]
 
-#[cfg(feature = "serde-serialize")]
-extern crate serde;
-#[cfg(feature = "serde-serialize")]
-extern crate serde_json;
-
-extern crate wasm_bindgen_macro;
-
 use core::fmt;
 use core::marker;
 use core::mem;
 use core::ops::{Deref, DerefMut};
-use core::ptr;
 
-use convert::FromWasmAbi;
+use crate::convert::{FromWasmAbi, WasmOptionalF64, WasmSlice};
 
 macro_rules! if_std {
     ($($i:item)*) => ($(
@@ -30,17 +22,38 @@ macro_rules! if_std {
     )*)
 }
 
+macro_rules! externs {
+    ($(#[$attr:meta])* extern "C" { $(fn $name:ident($($args:tt)*) -> $ret:ty;)* }) => (
+        #[cfg(all(target_arch = "wasm32", not(target_os = "emscripten")))]
+        $(#[$attr])*
+        extern "C" {
+            $(fn $name($($args)*) -> $ret;)*
+        }
+
+        $(
+            #[cfg(not(all(target_arch = "wasm32", not(target_os = "emscripten"))))]
+            #[allow(unused_variables)]
+            unsafe extern fn $name($($args)*) -> $ret {
+                panic!("function not implemented on non-wasm32 targets")
+            }
+        )*
+    )
+}
+
 /// A module which is typically glob imported from:
 ///
 /// ```
 /// use wasm_bindgen::prelude::*;
 /// ```
 pub mod prelude {
+    pub use crate::JsValue;
+    pub use crate::UnwrapThrowExt;
+    #[doc(hidden)]
+    pub use wasm_bindgen_macro::__wasm_bindgen_class_marker;
     pub use wasm_bindgen_macro::wasm_bindgen;
-    pub use JsValue;
 
     if_std! {
-        pub use closure::Closure;
+        pub use crate::closure::Closure;
     }
 }
 
@@ -48,12 +61,16 @@ pub mod convert;
 pub mod describe;
 
 mod cast;
-pub use cast::JsCast;
+pub use crate::cast::JsCast;
 
 if_std! {
     extern crate std;
     use std::prelude::v1::*;
     pub mod closure;
+    mod anyref;
+
+    mod cache;
+    pub use cache::intern::{intern, unintern};
 }
 
 /// Representation of an object owned by JS.
@@ -156,9 +173,13 @@ impl JsValue {
     /// JS object corresponding to the symbol created.
     pub fn symbol(description: Option<&str>) -> JsValue {
         unsafe {
-            let ptr = description.map(|s| s.as_ptr()).unwrap_or(ptr::null());
-            let len = description.map(|s| s.len()).unwrap_or(0);
-            JsValue::_new(__wbindgen_symbol_new(ptr, len))
+            match description {
+                Some(description) => JsValue::_new(__wbindgen_symbol_named_new(
+                    description.as_ptr(),
+                    description.len(),
+                )),
+                None => JsValue::_new(__wbindgen_symbol_anonymous_new()),
+            }
         }
     }
 
@@ -205,10 +226,8 @@ impl JsValue {
         T: for<'a> serde::de::Deserialize<'a>,
     {
         unsafe {
-            let mut ptr = ptr::null_mut();
-            let len = __wbindgen_json_serialize(self.idx, &mut ptr);
-            let s = Vec::from_raw_parts(ptr, len, len);
-            let s = String::from_utf8_unchecked(s);
+            let ret = __wbindgen_json_serialize(self.idx);
+            let s = String::from_abi(ret);
             serde_json::from_str(&s)
         }
     }
@@ -219,15 +238,7 @@ impl JsValue {
     /// If this JS value is not an instance of a number then this returns
     /// `None`.
     pub fn as_f64(&self) -> Option<f64> {
-        let mut invalid = 0;
-        unsafe {
-            let ret = __wbindgen_number_get(self.idx, &mut invalid);
-            if invalid == 1 {
-                None
-            } else {
-                Some(ret)
-            }
-        }
+        unsafe { FromWasmAbi::from_abi(__wbindgen_number_get(self.idx)) }
     }
 
     /// Tests whether this JS value is a JS string.
@@ -245,18 +256,19 @@ impl JsValue {
     ///
     /// If this JS value is not an instance of a string or if it's not valid
     /// utf-8 then this returns `None`.
+    ///
+    /// # UTF-16 vs UTF-8
+    ///
+    /// JavaScript strings in general are encoded as UTF-16, but Rust strings
+    /// are encoded as UTF-8. This can cause the Rust string to look a bit
+    /// different than the JS string sometimes. For more details see the
+    /// [documentation about the `str` type][caveats] which contains a few
+    /// caveats about the encodings.
+    ///
+    /// [caveats]: https://rustwasm.github.io/docs/wasm-bindgen/reference/types/str.html
     #[cfg(feature = "std")]
     pub fn as_string(&self) -> Option<String> {
-        unsafe {
-            let mut len = 0;
-            let ptr = __wbindgen_string_get(self.idx, &mut len);
-            if ptr.is_null() {
-                None
-            } else {
-                let data = Vec::from_raw_parts(ptr, len, len);
-                Some(String::from_utf8_unchecked(data))
-            }
-        }
+        unsafe { FromWasmAbi::from_abi(__wbindgen_string_get(self.idx)) }
     }
 
     /// Returns the `bool` value of this JS value if it's an instance of a
@@ -304,18 +316,30 @@ impl JsValue {
         unsafe { __wbindgen_is_function(self.idx) == 1 }
     }
 
+    /// Tests whether the value is ["truthy"].
+    ///
+    /// ["truthy"]: https://developer.mozilla.org/en-US/docs/Glossary/Truthy
+    #[inline]
+    pub fn is_truthy(&self) -> bool {
+        !self.is_falsy()
+    }
+
+    /// Tests whether the value is ["falsy"].
+    ///
+    /// ["falsy"]: https://developer.mozilla.org/en-US/docs/Glossary/Falsy
+    #[inline]
+    pub fn is_falsy(&self) -> bool {
+        unsafe { __wbindgen_is_falsy(self.idx) == 1 }
+    }
+
     /// Get a string representation of the JavaScript object for debugging
     #[cfg(feature = "std")]
     fn as_debug_string(&self) -> String {
         unsafe {
-            let mut len = 0;
-            let ptr = __wbindgen_debug_string(self.idx, &mut len);
-            if ptr.is_null() {
-                unreachable!("`__wbindgen_debug_string` must return a valid string")
-            } else {
-                let data = Vec::from_raw_parts(ptr, len, len);
-                String::from_utf8_unchecked(data)
-            }
+            let mut ret = [0; 2];
+            __wbindgen_debug_string(&mut ret, self.idx);
+            let data = Vec::from_raw_parts(ret[0] as *mut u8, ret[1], ret[1]);
+            String::from_utf8_unchecked(data)
         }
     }
 }
@@ -459,58 +483,54 @@ macro_rules! numbers {
 
 numbers! { i8 u8 i16 u16 i32 u32 f32 f64 }
 
-macro_rules! externs {
-    ($(fn $name:ident($($args:tt)*) -> $ret:ty;)*) => (
-        #[cfg(all(target_arch = "wasm32", not(target_os = "emscripten")))]
-        #[link(wasm_import_module = "__wbindgen_placeholder__")]
-        extern "C" {
-            $(fn $name($($args)*) -> $ret;)*
-        }
-
-        $(
-            #[cfg(not(all(target_arch = "wasm32", not(target_os = "emscripten"))))]
-            #[allow(unused_variables)]
-            unsafe extern "C" fn $name($($args)*) -> $ret {
-                panic!("function not implemented on non-wasm32 targets")
-            }
-        )*
-    )
-}
-
 externs! {
-    fn __wbindgen_object_clone_ref(idx: u32) -> u32;
-    fn __wbindgen_object_drop_ref(idx: u32) -> ();
-    fn __wbindgen_string_new(ptr: *const u8, len: usize) -> u32;
-    fn __wbindgen_number_new(f: f64) -> u32;
-    fn __wbindgen_number_get(idx: u32, invalid: *mut u8) -> f64;
-    fn __wbindgen_is_null(idx: u32) -> u32;
-    fn __wbindgen_is_undefined(idx: u32) -> u32;
-    fn __wbindgen_boolean_get(idx: u32) -> u32;
-    fn __wbindgen_symbol_new(ptr: *const u8, len: usize) -> u32;
-    fn __wbindgen_is_symbol(idx: u32) -> u32;
-    fn __wbindgen_is_object(idx: u32) -> u32;
-    fn __wbindgen_is_function(idx: u32) -> u32;
-    fn __wbindgen_is_string(idx: u32) -> u32;
-    fn __wbindgen_string_get(idx: u32, len: *mut usize) -> *mut u8;
-    fn __wbindgen_debug_string(idx: u32, len: *mut usize) -> *mut u8;
-    fn __wbindgen_throw(a: *const u8, b: usize) -> !;
-    fn __wbindgen_rethrow(a: u32) -> !;
+    #[link(wasm_import_module = "__wbindgen_placeholder__")]
+    extern "C" {
+        fn __wbindgen_object_clone_ref(idx: u32) -> u32;
+        fn __wbindgen_object_drop_ref(idx: u32) -> ();
 
-    fn __wbindgen_cb_drop(idx: u32) -> u32;
-    fn __wbindgen_cb_forget(idx: u32) -> ();
+        fn __wbindgen_string_new(ptr: *const u8, len: usize) -> u32;
+        fn __wbindgen_number_new(f: f64) -> u32;
+        fn __wbindgen_symbol_named_new(ptr: *const u8, len: usize) -> u32;
+        fn __wbindgen_symbol_anonymous_new() -> u32;
 
-    fn __wbindgen_describe(v: u32) -> ();
-    fn __wbindgen_describe_closure(a: u32, b: u32, c: u32) -> u32;
+        fn __wbindgen_anyref_heap_live_count() -> u32;
 
-    fn __wbindgen_json_parse(ptr: *const u8, len: usize) -> u32;
-    fn __wbindgen_json_serialize(idx: u32, ptr: *mut *mut u8) -> usize;
-    fn __wbindgen_jsval_eq(a: u32, b: u32) -> u32;
+        fn __wbindgen_is_null(idx: u32) -> u32;
+        fn __wbindgen_is_undefined(idx: u32) -> u32;
+        fn __wbindgen_is_symbol(idx: u32) -> u32;
+        fn __wbindgen_is_object(idx: u32) -> u32;
+        fn __wbindgen_is_function(idx: u32) -> u32;
+        fn __wbindgen_is_string(idx: u32) -> u32;
+        fn __wbindgen_is_falsy(idx: u32) -> u32;
 
-    fn __wbindgen_memory() -> u32;
-    fn __wbindgen_module() -> u32;
+        fn __wbindgen_number_get(idx: u32) -> WasmOptionalF64;
+        fn __wbindgen_boolean_get(idx: u32) -> u32;
+        fn __wbindgen_string_get(idx: u32) -> WasmSlice;
+
+        fn __wbindgen_debug_string(ret: *mut [usize; 2], idx: u32) -> ();
+
+        fn __wbindgen_throw(a: *const u8, b: usize) -> !;
+        fn __wbindgen_rethrow(a: u32) -> !;
+
+        fn __wbindgen_cb_drop(idx: u32) -> u32;
+        fn __wbindgen_cb_forget(idx: u32) -> ();
+
+        fn __wbindgen_describe(v: u32) -> ();
+        fn __wbindgen_describe_closure(a: u32, b: u32, c: u32) -> u32;
+
+        fn __wbindgen_json_parse(ptr: *const u8, len: usize) -> u32;
+        fn __wbindgen_json_serialize(idx: u32) -> WasmSlice;
+        fn __wbindgen_jsval_eq(a: u32, b: u32) -> u32;
+
+        fn __wbindgen_memory() -> u32;
+        fn __wbindgen_module() -> u32;
+        fn __wbindgen_function_table() -> u32;
+    }
 }
 
 impl Clone for JsValue {
+    #[inline]
     fn clone(&self) -> JsValue {
         unsafe {
             let idx = __wbindgen_object_clone_ref(self.idx);
@@ -538,7 +558,7 @@ impl Drop for JsValue {
     fn drop(&mut self) {
         unsafe {
             // We definitely should never drop anything in the stack area
-            debug_assert!(self.idx >= JSIDX_OFFSET);
+            debug_assert!(self.idx >= JSIDX_OFFSET, "free of stack slot {}", self.idx);
 
             // Otherwise if we're not dropping one of our reserved values,
             // actually call the intrinsic. See #1054 for eventually removing
@@ -638,10 +658,129 @@ pub fn throw_val(s: JsValue) -> ! {
     }
 }
 
+/// Get the count of live `anyref`s / `JsValue`s in `wasm-bindgen`'s heap.
+///
+/// ## Usage
+///
+/// This is intended for debugging and writing tests.
+///
+/// To write a test that asserts against unnecessarily keeping `anref`s /
+/// `JsValue`s alive:
+///
+/// * get an initial live count,
+///
+/// * perform some series of operations or function calls that should clean up
+///   after themselves, and should not keep holding onto `anyref`s / `JsValue`s
+///   after completion,
+///
+/// * get the final live count,
+///
+/// * and assert that the initial and final counts are the same.
+///
+/// ## What is Counted
+///
+/// Note that this only counts the *owned* `anyref`s / `JsValue`s that end up in
+/// `wasm-bindgen`'s heap. It does not count borrowed `anyref`s / `JsValue`s
+/// that are on its stack.
+///
+/// For example, these `JsValue`s are accounted for:
+///
+/// ```ignore
+/// #[wasm_bindgen]
+/// pub fn my_function(this_is_counted: JsValue) {
+///     let also_counted = JsValue::from_str("hi");
+///     assert!(wasm_bindgen::anyref_heap_live_count() >= 2);
+/// }
+/// ```
+///
+/// While this borrowed `JsValue` ends up on the stack, not the heap, and
+/// therefore is not accounted for:
+///
+/// ```ignore
+/// #[wasm_bindgen]
+/// pub fn my_other_function(this_is_not_counted: &JsValue) {
+///     // ...
+/// }
+/// ```
+pub fn anyref_heap_live_count() -> u32 {
+    unsafe { __wbindgen_anyref_heap_live_count() }
+}
+
+/// An extension trait for `Option<T>` and `Result<T, E>` for unwraping the `T`
+/// value, or throwing a JS error if it is not available.
+///
+/// These methods should have a smaller code size footprint than the normal
+/// `Option::unwrap` and `Option::expect` methods, but they are specific to
+/// working with wasm and JS.
+///
+/// On non-wasm32 targets, defaults to the normal unwrap/expect calls.
+///
+/// # Example
+///
+/// ```
+/// use wasm_bindgen::prelude::*;
+///
+/// // If the value is `Option::Some` or `Result::Ok`, then we just get the
+/// // contained `T` value.
+/// let x = Some(42);
+/// assert_eq!(x.unwrap_throw(), 42);
+///
+/// let y: Option<i32> = None;
+///
+/// // This call would throw an error to JS!
+/// //
+/// //     y.unwrap_throw()
+/// //
+/// // And this call would throw an error to JS with a custom error message!
+/// //
+/// //     y.expect_throw("woopsie daisy!")
+/// ```
+pub trait UnwrapThrowExt<T>: Sized {
+    /// Unwrap this `Option` or `Result`, but instead of panicking on failure,
+    /// throw an exception to JavaScript.
+    fn unwrap_throw(self) -> T {
+        self.expect_throw("`unwrap_throw` failed")
+    }
+
+    /// Unwrap this container's `T` value, or throw an error to JS with the
+    /// given message if the `T` value is unavailable (e.g. an `Option<T>` is
+    /// `None`).
+    fn expect_throw(self, message: &str) -> T;
+}
+
+impl<T> UnwrapThrowExt<T> for Option<T> {
+    fn expect_throw(self, message: &str) -> T {
+        if cfg!(all(target_arch = "wasm32", not(target_os = "emscripten"))) {
+            match self {
+                Some(val) => val,
+                None => throw_str(message),
+            }
+        } else {
+            self.expect(message)
+        }
+    }
+}
+
+impl<T, E> UnwrapThrowExt<T> for Result<T, E>
+where
+    E: core::fmt::Debug,
+{
+    fn expect_throw(self, message: &str) -> T {
+        if cfg!(all(target_arch = "wasm32", not(target_os = "emscripten"))) {
+            match self {
+                Ok(val) => val,
+                Err(_) => throw_str(message),
+            }
+        } else {
+            self.expect(message)
+        }
+    }
+}
+
 /// Returns a handle to this wasm instance's `WebAssembly.Module`
 ///
 /// Note that this is only available when the final wasm app is built with
-/// `--no-modules`, it's not recommended to rely on this API yet! This is
+/// `--target no-modules`, it's not recommended to rely on this API yet! This is
 /// largely just an experimental addition to enable threading demos. Using this
 /// may prevent your wasm module from building down the road.
 #[doc(hidden)]
@@ -654,10 +793,18 @@ pub fn memory() -> JsValue {
     unsafe { JsValue::_new(__wbindgen_memory()) }
 }
 
+/// Returns a handle to this wasm instance's `WebAssembly.Table` which is the
+/// indirect function table used by Rust
+pub fn function_table() -> JsValue {
+    unsafe { JsValue::_new(__wbindgen_function_table()) }
+}
+
 #[doc(hidden)]
 pub mod __rt {
+    use crate::JsValue;
     use core::cell::{Cell, UnsafeCell};
     use core::ops::{Deref, DerefMut};
+
     pub extern crate core;
     #[cfg(feature = "std")]
     pub extern crate std;
@@ -815,7 +962,7 @@ pub mod __rt {
     }
 
     if_std! {
-        use std::alloc::{alloc, dealloc, Layout};
+        use std::alloc::{alloc, dealloc, realloc, Layout};
         use std::mem;
 
         #[no_mangle]
@@ -834,6 +981,25 @@ pub mod __rt {
                 }
             }
 
+            malloc_failure();
+        }
+
+        #[no_mangle]
+        pub unsafe extern "C" fn __wbindgen_realloc(ptr: *mut u8, old_size: usize, new_size: usize) -> *mut u8 {
+            let align = mem::align_of::<usize>();
+            debug_assert!(old_size > 0);
+            debug_assert!(new_size > 0);
+            if let Ok(layout) = Layout::from_size_align(old_size, align) {
+                let ptr = realloc(ptr, layout, new_size);
+                if !ptr.is_null() {
+                    return ptr
+                }
+            }
+            malloc_failure();
+        }
+
+        #[cold]
+        fn malloc_failure() -> ! {
             if cfg!(debug_assertions) {
                 super::throw_str("invalid malloc request")
             } else {
@@ -854,20 +1020,6 @@ pub mod __rt {
         }
     }
 
-    pub const GLOBAL_STACK_CAP: usize = 16;
-
-    // Increase the alignment to 8 here because this can be used as a
-    // BigUint64Array pointer base which requires alignment 8
-    #[repr(align(8))]
-    struct GlobalData([u32; GLOBAL_STACK_CAP]);
-
-    static mut GLOBAL_STACK: GlobalData = GlobalData([0; GLOBAL_STACK_CAP]);
-
-    #[no_mangle]
-    pub unsafe extern "C" fn __wbindgen_global_argument_ptr() -> *mut u32 {
-        GLOBAL_STACK.0.as_mut_ptr()
-    }
-
     /// This is a curious function necessary to get wasm-bindgen working today,
     /// and it's a bit of an unfortunate hack.
     ///
@@ -901,7 +1053,89 @@ pub mod __rt {
     /// in the object file and link the intrinsics.
     ///
     /// Ideas for how to improve this are most welcome!
-    pub fn link_mem_intrinsics() {}
+    pub fn link_mem_intrinsics() {
+        crate::anyref::link_intrinsics();
+    }
+
+    static mut GLOBAL_EXNDATA: [u32; 2] = [0; 2];
+
+    #[no_mangle]
+    pub unsafe extern "C" fn __wbindgen_exn_store(idx: u32) {
+        debug_assert_eq!(GLOBAL_EXNDATA[0], 0);
+        GLOBAL_EXNDATA[0] = 1;
+        GLOBAL_EXNDATA[1] = idx;
+    }
+
+    pub fn take_last_exception() -> Result<(), super::JsValue> {
+        unsafe {
+            let ret = if GLOBAL_EXNDATA[0] == 1 {
+                Err(super::JsValue::_new(GLOBAL_EXNDATA[1]))
+            } else {
+                Ok(())
+            };
+            GLOBAL_EXNDATA[0] = 0;
+            GLOBAL_EXNDATA[1] = 0;
+            return ret;
+        }
+    }
+
+    /// An internal helper trait for usage in `#[wasm_bindgen]` on `async`
+    /// functions to convert the return value of the function to
+    /// `Result<JsValue, JsValue>` which is what we'll return to JS (where an
+    /// error is a failed future).
+    pub trait IntoJsResult {
+        fn into_js_result(self) -> Result<JsValue, JsValue>;
+    }
+
+    impl IntoJsResult for () {
+        fn into_js_result(self) -> Result<JsValue, JsValue> {
+            Ok(JsValue::undefined())
+        }
+    }
+
+    impl<T: Into<JsValue>> IntoJsResult for T {
+        fn into_js_result(self) -> Result<JsValue, JsValue> {
+            Ok(self.into())
+        }
+    }
+
+    impl<T: Into<JsValue>, E: Into<JsValue>> IntoJsResult for Result<T, E> {
+        fn into_js_result(self) -> Result<JsValue, JsValue> {
+            match self {
+                Ok(e) => Ok(e.into()),
+                Err(e) => Err(e.into()),
+            }
+        }
+    }
+
+    impl<E: Into<JsValue>> IntoJsResult for Result<(), E> {
+        fn into_js_result(self) -> Result<JsValue, JsValue> {
+            match self {
+                Ok(()) => Ok(JsValue::undefined()),
+                Err(e) => Err(e.into()),
+            }
+        }
+    }
+
+    /// An internal helper trait for usage in `#[wasm_bindgen(start)]`
+    /// functions to throw the error (if it is `Err`).
+    pub trait Start {
+        fn start(self);
+    }
+
+    impl Start for () {
+        #[inline]
+        fn start(self) {}
+    }
+
+    impl<E: Into<JsValue>> Start for Result<(), E> {
+        #[inline]
+        fn start(self) {
+            if let Err(e) = self {
+                crate::throw_val(e.into());
+            }
+        }
+    }
 }
 
 /// A wrapper type around slices and vectors for binding the `Uint8ClampedArray`
